package dev.kske.chess.event;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

/**
 * Project: <strong>Chess</strong><br>
 * File: <strong>EventBus.java</strong><br>
 * Created: <strong>02.08.2019</strong><br>
 * Author: <strong>Kai S. K. Engelbart</strong>
 */
public final class EventBus {

	private Map<Class<?>, List<Invocation>> invocations;

	private static EventBus instance;

	public static EventBus getInstance() {
		if (instance == null) instance = new EventBus();
		return instance;
	}

	private EventBus() {
		invocations = new ConcurrentHashMap<>();
	}

	public void post(Object object) {
		Class<?> type = object.getClass();
		if (invocations.containsKey(type)) invocations.get(type).forEach(invocation -> invocation.invoke(object));
	}

	public void register(Object object) {
		Class<?> currentClass = object.getClass();
		while (currentClass != null) {
			for (Method method : findSubscriptionMethods(currentClass)) {
				Class<?> type = method.getParameterTypes()[0];
				if (invocations.containsKey(type)) invocations.get(type).add(new Invocation(method, object));
				else {
					List<Invocation> temp = new Vector<>();
					temp.add(new Invocation(method, object));
					invocations.put(type, temp);
				}
			}
			currentClass = currentClass.getSuperclass();
		}
	}

	private List<Method> findSubscriptionMethods(Class<?> type) {
		List<Method> subscriberMethods = Arrays.stream(type.getDeclaredMethods())
			.filter(method -> method.isAnnotationPresent(Subscribe.class))
			.collect(Collectors.toList());
		checkSubscriberMethods(subscriberMethods);
		return subscriberMethods;
	}

	private void checkSubscriberMethods(List<Method> subscriberMethods) {
		if (subscriberMethods.stream().anyMatch(method -> method.getParameterCount() != 1))
			throw new IllegalArgumentException("Method annotated with @Subscribe must have exactly one parameter");
	}
}
